#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
import idaes.core.util.scaling as iscale
from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock
from watertap.unit_models.boron_removal import BoronRemoval
from pyomo.environ import (
    ConcreteModel,
    units as pyunits,
)
from idaes.core import (
    FlowsheetBlock,
)
from idaes.core.util.exceptions import ConfigurationError
from watertap.unit_models.tests.unit_test_harness import UnitTestHarness
from watertap.core.solvers import get_solver
import re

__author__ = "Austin Ladshaw"

solver = get_solver()


# Helper function for multiple test setup
def model_setup(
    m,
    chem_add=1.8e-5,
    state={
        "H2O": 100,
        "H_+": 1e-7,
        "OH_-": 1e-7,
        "B[OH]3": 2e-4,
        "B[OH]4_-": 1e-6,
        "Na_+": 1e-4,
        "HCO3_-": 1e-4,
    },
):

    m.fs.unit.inlet.pressure.fix(101325)
    m.fs.unit.inlet.temperature.fix(298.15)
    for j in state:
        idx = (0, "Liq", j)
        if idx in m.fs.unit.inlet.flow_mol_phase_comp:
            m.fs.unit.inlet.flow_mol_phase_comp[idx].fix(state[j])
    m.fs.unit.caustic_dose_rate.fix(chem_add)
    m.fs.unit.reactor_volume.fix(1)


# Helper function to automate scaling
def scaling_setup(
    m,
    state={
        "H2O": 100,
        "H_+": 5e-5,
        "OH_-": 5e-5,
        "B[OH]3": 2e-4,
        "B[OH]4_-": 1e-6,
        "Na_+": 1e-4,
        "HCO3_-": 1e-4,
    },
):
    # Set some scaling factors and look for 'bad' scaling
    for j in state:
        idx = (0, "Liq", j)
        if idx in m.fs.unit.inlet.flow_mol_phase_comp:
            m.fs.properties.set_default_scaling(
                "flow_mol_phase_comp", 1 / state[j], index=("Liq", j)
            )

    iscale.calculate_scaling_factors(m.fs)


# -----------------------------------------------------------------------------
# Start test class
def min_boron_removal_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    # create dict to define ions (the prop pack requires this)
    ion_dict = {
        "solute_list": ["B[OH]3", "B[OH]4_-", "Na_+"],
        "mw_data": {
            "H2O": 18e-3,
            "B[OH]3": 61.83e-3,
            "B[OH]4_-": 78.83e-3,
            "Na_+": 23e-3,
        },
        "charge": {
            "B[OH]4_-": -1,
            "Na_+": 1,
            "B[OH]3": 0,
        },
    }

    # attach prop pack to flowsheet
    m.fs.properties = MCASParameterBlock(**ion_dict)

    map = {
        "boron_name": "B[OH]3",  # [is required]
        "borate_name": "B[OH]4_-",  # [is required]
        "caustic_additive": {
            "cation_name": "Na_+",  # [is optional]
            "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
            "moles_cation_per_additive": 1,  # [is required]
        },
    }
    m.fs.unit = BoronRemoval(
        property_package=m.fs.properties, chemical_mapping_data=map
    )
    state = {
        "H2O": 100,
        "H_+": 1e-7,
        "OH_-": 1e-7,
        "B[OH]3": 2e-4,
        "B[OH]4_-": 1e-6,
        "Na_+": 1e-5,
        "HCO3_-": 1e-4,
    }

    # For this test, we don't actually know how much
    #   of the base to add. Instead, we have a target
    #   exit flow of boron (see below). We will initially
    #   guess that we want 5 mg/L of additive, but the
    #   actual solution is 10 mg/L.
    model_setup(m, chem_add=0.9e-5, state=state)

    # Modified this test to fix the desired outlet flow
    #   of Boron and unfix the dosage needed to get that outlet
    m.fs.unit.outlet.flow_mol_phase_comp[(0, "Liq", "B[OH]3")].fix(1.98677e-5)
    m.fs.unit.caustic_dose_rate.unfix()

    scaling_setup(m)

    return m


class TestBoronRemoval_IonPropPack_Min(UnitTestHarness):
    def configure(self):
        m = min_boron_removal_model()

        self.unit_solutions[
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"]
        ] = 1.98677e-5
        self.unit_solutions[
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", "B[OH]4_-"]
        ] = 1.81133e-4
        self.unit_solutions[m.fs.unit.outlet_pH()] = 10.171
        self.unit_solutions[m.fs.unit.outlet_pOH()] = 3.8257
        self.unit_solutions[m.fs.unit.caustic_dose_rate[0]] = 1.8e-5

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"]
                + m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"],
                "out": m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"]
                + m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"],
            },
        }

        return m


def alk_boron_removal_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # create dict to define ions (the prop pack requires this)
    ion_dict = {
        "solute_list": ["B[OH]3", "B[OH]4_-", "Na_+", "HCO3_-"],
        "mw_data": {
            "H2O": 18e-3,
            "B[OH]3": 61.83e-3,
            "B[OH]4_-": 78.83e-3,
            "Na_+": 23e-3,
            "HCO3_-": 61e-3,
        },
        "charge": {"B[OH]3": 0, "B[OH]4_-": -1, "Na_+": 1, "HCO3_-": -1},
    }

    # attach prop pack to flowsheet
    m.fs.properties = MCASParameterBlock(**ion_dict)

    map = {
        "boron_name": "B[OH]3",  # [is required]
        "borate_name": "B[OH]4_-",  # [is required]
        "caustic_additive": {
            "cation_name": "Na_+",  # [is optional]
            "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
            "moles_cation_per_additive": 1,  # [is required]
        },
    }
    m.fs.unit = BoronRemoval(
        property_package=m.fs.properties, chemical_mapping_data=map
    )

    model_setup(m, chem_add=1.8e-4)
    scaling_setup(m)

    return m


class TestBoronRemoval_IonPropPack_with_ResAlk(UnitTestHarness):
    def configure(self):
        m = alk_boron_removal_model()

        self.unit_solutions[
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"]
        ] = 1.36911e-6
        self.unit_solutions[
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", "B[OH]4_-"]
        ] = 1.99642e-4
        self.unit_solutions[m.fs.unit.outlet_pH()] = 11.375
        self.unit_solutions[m.fs.unit.outlet_pOH()] = 2.6218

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"]
                + m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"],
                "out": m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"]
                + m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"],
            },
        }

        return m


def base_boron_removal_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # create dict to define ions (the prop pack requires this)
    ion_dict = {
        "solute_list": ["B[OH]3", "B[OH]4_-", "Na_+"],
        "mw_data": {
            "H2O": 18e-3,
            "B[OH]3": 61.83e-3,
            "B[OH]4_-": 78.83e-3,
            "Na_+": 23e-3,
        },
        "charge": {
            "B[OH]3": 0,
            "B[OH]4_-": -1,
            "Na_+": 1,
        },
    }

    # attach prop pack to flowsheet
    m.fs.properties = MCASParameterBlock(**ion_dict)

    map = {
        "boron_name": "B[OH]3",  # [is required]
        "borate_name": "B[OH]4_-",  # [is required]
        "caustic_additive": {
            "cation_name": "Na_+",  # [is optional]
            "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
            "moles_cation_per_additive": 1,  # [is required]
        },
    }
    m.fs.unit = BoronRemoval(
        property_package=m.fs.properties, chemical_mapping_data=map
    )

    model_setup(m, chem_add=0)
    scaling_setup(m)

    return m


class TestBoronRemoval_IonPropPack_with_ResBase(UnitTestHarness):
    def configure(self):
        m = base_boron_removal_model()

        self.unit_solutions[
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"]
        ] = 1.20642e-4
        self.unit_solutions[
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", "B[OH]4_-"]
        ] = 8.03579e-5
        self.unit_solutions[m.fs.unit.outlet_pH()] = 9.0343
        self.unit_solutions[m.fs.unit.outlet_pOH()] = 4.9621

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"]
                + m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"],
                "out": m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"]
                + m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "B[OH]3"],
            },
        }

        return m


# -----------------------------------------------------------------------------
# Start test class with bad config
class TestBoronRemoval_BadConfigs:
    @pytest.fixture(scope="class")
    def boron_removal_bad_configs(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # create dict to define ions (the prop pack requires this)
        ion_dict = {
            "solute_list": ["B[OH]3", "B[OH]4_-", "H_+", "OH_-", "Na_+"],
            "mw_data": {
                "H2O": 18e-3,
                "B[OH]3": 61.83e-3,
                "B[OH]4_-": 78.83e-3,
                "H_+": 1e-3,
                "OH_-": 17e-3,
                "Na_+": 23e-3,
            },
            "charge": {
                "B[OH]4_-": -1,
            },
        }

        # attach prop pack to flowsheet
        m.fs.properties = MCASParameterBlock(**ion_dict, ignore_neutral_charge=True)

        return m

    @pytest.mark.unit
    def test_build_failures(self, boron_removal_bad_configs):
        m = boron_removal_bad_configs

        # Invalid name of boron
        map = {
            "boron_name": "Boron",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'boron_name' {Boron} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Invalid name of borate
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "Borate",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'borate_name' {Borate} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Empty dict
        map = {}
        error_msg = "Did not provide a 'dict' for 'chemical_mapping_data' "
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Invalid name of hydroxide
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "hydroxide_name": "Raccoon_City",
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'hydroxide_name' {Raccoon_City} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Invalid name of protons
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "proton_name": "Pulled_Pork",
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'proton_name' {Pulled_Pork} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Invalid name of cation
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Im_Batman",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'cation_name' {Im_Batman} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Missing information (borate_name)
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "caustic_additive": {
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = "Missing some required information in 'chemical_mapping_data' "
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Missing information (mw_additive)
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = "Missing some required information in 'chemical_mapping_data' "
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Improper data type (mw_additive must be a tuple with value and units)
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is required]
                "mw_additive": 40,  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = "Did not provide a tuple for 'mw_additive' "
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Thermophysical property package to be used in conjunction with modified ASM2d reactions.

Reference:
[1] X. Flores-Alsina, K. Solon, C.K. Mbamba, S. Tait, K.V. Gernaey, U. Jeppsson, D.J. Batstone,
Modelling phosphorus (P), sulfur (S) and iron (Fe) interactions for dynamic simulations of anaerobic digestion processes,
Water Research. 95 (2016) 370-382. https://www.sciencedirect.com/science/article/pii/S0043135416301397

[2] K. Solon, X. Flores-Alsina, C. Kazadi Mbamba, D. Ikumi, E.I.P. Volcke, C. Vaneeckhaute, G. Ekama,
P.A. Vanrolleghem, D.J. Batstone, K.V. Gernaey, U. Jeppsson, Plant-wide modelling of phosphorus transformations in
wastewater treatment systems: Impacts of control and operational strategies, Water Research. 113 (2017) 97-110
https://www.sciencedirect.com/science/article/pii/S0043135417300829

"""

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    LiquidPhase,
    Solute,
    Solvent,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
import idaes.logger as idaeslog
from idaes.core.scaling import CustomScalerBase
from idaes.core.base.property_base import PhysicalParameterBlock

# Some more information about this module
__author__ = "Marcus Holly, Adam Atia, Xinhong Liu"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ModifiedASM2dParameterBlock")
class ModifiedASM2dParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class
    """

    def build(self):
        """
        Callable method for Block construction.
        """
        super().build()

        self._state_block_class = ModifiedASM2dStateBlock

        # Add Phase objects
        self.Liq = LiquidPhase()

        # Add Component objects
        self.H2O = Solvent()

        # Soluble species
        self.S_A = Solute(
            doc="Fermentation products, considered to be acetate. [kg COD/m^3]"
        )
        self.S_F = Solute(
            doc="Fermentable, readily bio-degradable organic substrates. [kg COD/m^3]"
        )
        self.S_I = Solute(doc="Inert soluble organic material. [kg COD/m^3]")
        self.S_N2 = Solute(
            doc="Dinitrogen, N2. SN2 is assumed to be the only nitrogenous product of denitrification. [kg N/m^3]"
        )
        self.S_NH4 = Solute(doc="Ammonium plus ammonia nitrogen. [kg N/m^3]")
        self.S_NO3 = Solute(
            doc="Nitrate plus nitrite nitrogen (N03' + N02' -N). SN03 is assumed to include nitrate as well as nitrite nitrogen. [kg N/m^3]"
        )
        self.S_O2 = Solute(doc="Dissolved oxygen. [kg O2/m^3]")
        self.S_PO4 = Solute(
            doc="Inorganic soluble phosphorus, primarily ortho-phosphates. [kg P/m^3]"
        )
        self.S_K = Solute(doc="Potassium, [kg K/m^3]")
        self.S_Mg = Solute(doc="Magnesium, [kg Mg/m^3]")
        self.S_IC = Solute(doc="Inorganic carbon, [kg C/m^3]")

        # Particulate species
        self.X_AUT = Solute(doc="Autotrophic nitrifying organisms. [kg COD/m^3]")
        self.X_H = Solute(doc="Heterotrophic organisms. [kg COD/m^3]")
        self.X_I = Solute(doc="Inert particulate organic material. [kg COD/m^3]")
        self.X_PAO = Solute(doc="Phosphate-accumulating organisms. [kg COD/m^3]")
        self.X_PHA = Solute(
            doc="A cell internal storage product of phosphorus-accumulating organisms, primarily comprising poly-hydroxy-alkanoates (PHA). [kg COD/m^3]"
        )
        self.X_PP = Solute(doc="Poly-phosphate. [kg P/m^3]")
        self.X_S = Solute(doc="Slowly biodegradable substrates. [kg COD/m^3]")

        # Create sets for use across ASM models and associated unit models
        self.non_particulate_component_set = pyo.Set(
            initialize=[
                "S_A",
                "S_F",
                "S_I",
                "S_N2",
                "S_NH4",
                "S_NO3",
                "S_O2",
                "S_PO4",
                "S_K",
                "S_Mg",
                "S_IC",
                "H2O",
            ]
        )
        self.particulate_component_set = pyo.Set(
            initialize=[
                "X_AUT",
                "X_H",
                "X_I",
                "X_PAO",
                "X_PHA",
                "X_PP",
                "X_S",
            ]
        )
        self.tss_component_set = pyo.Set(
            initialize=[
                "X_AUT",
                "X_H",
                "X_I",
                "X_PAO",
                "X_PHA",
                "X_PP",
                "X_S",
            ]
        )

        # Heat capacity of water
        self.cp_mass = pyo.Param(
            initialize=4182,
            doc="Specific heat capacity of water",
            units=pyo.units.J / pyo.units.kg / pyo.units.K,
        )
        # Density of water
        self.dens_mass = pyo.Param(
            initialize=997,
            doc="Density of water",
            units=pyo.units.kg / pyo.units.m**3,
        )

        # Thermodynamic reference state
        self.pressure_ref = pyo.Param(
            within=pyo.PositiveReals,
            mutable=True,
            default=101325.0,
            doc="Reference pressure",
            units=pyo.units.Pa,
        )
        self.temperature_ref = pyo.Param(
            within=pyo.PositiveReals,
            mutable=True,
            default=298.15,
            doc="Reference temperature",
            units=pyo.units.K,
        )

        # COD to VSS coefficients
        self.CODtoVSS_XI = pyo.Var(
            initialize=1.5686,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="mass COD per mass VSS of XI",
        )
        self.CODtoVSS_XS = pyo.Var(
            initialize=1.5686,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="mass COD per mass VSS of XS",
        )
        self.CODtoVSS_XBM = pyo.Var(
            initialize=1.3072,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="mass COD per mass VSS of biomass",
        )
        self.CODtoVSS_XPHA = pyo.Var(
            initialize=1.9608,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="mass COD per mass VSS of XPHA",
        )
        # Inorganic solids parameters
        self.ISS_P = pyo.Var(
            initialize=3.23,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="mass ISS per mass P",
        )
        self.f_ISS_BM = pyo.Var(
            initialize=0.15,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="ISS fractional content of biomass",
        )

        # Effluent Quality Index (EQI) parameters [2]
        self.i_NSF = pyo.Var(
            initialize=0.03352,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="N content of fermentable substrate, S_F, [kg N/kg COD]",
        )
        self.i_NSI = pyo.Var(
            initialize=0.06003,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="N content of inert soluble COD S_I, [kg N/kg COD]",
        )
        self.i_NXI = pyo.Var(
            initialize=0.06003,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="N content of inert particulate COD X_I, [kg N/kg COD]",
        )
        self.i_NXS = pyo.Var(
            initialize=0.03352,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="N content of slowly biodegradable substrate X_S, [kg N/kg COD]",
        )
        self.i_NBM = pyo.Var(
            initialize=0.08615,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="N content of biomass, X_H, X_PAO, X_AUT, [kg N/kg COD]",
        )
        self.f_SI = pyo.Var(
            initialize=0.00,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="Production of S_I in hydrolysis, [kg COD/kg COD]",
        )
        self.f_XIH = pyo.Var(
            initialize=0.1,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="Fraction of inert COD generated in lysis of X_H, [kg COD/kg COD]",
        )
        self.f_XIP = pyo.Var(
            initialize=0.1,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="Fraction of inert COD generated in lysis of X_PAO and X_PHA, [kg COD/kg COD]",
        )
        self.f_XIA = pyo.Var(
            initialize=0.1,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="Fraction of inert COD generated in lysis of X_AUT, [kg COD/kg COD]",
        )
        self.i_PSF = pyo.Var(
            initialize=0.00559,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="P content of fermentable substrate, S_F, [kg P/kg COD]",
        )
        self.i_PSI = pyo.Var(
            initialize=0.00649,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="P content of inert soluble COD S_I, [kg P/kg COD]",
        )
        self.i_PXI = pyo.Var(
            initialize=0.00649,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="P content of inert particulate COD X_I, [kg P/kg COD]",
        )
        self.i_PXS = pyo.Var(
            initialize=0.00559,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="P content of slowly biodegradable substrate X_S, [kg P/kg COD]",
        )
        self.i_PBM = pyo.Var(
            initialize=0.02154,
            units=pyo.units.dimensionless,
            domain=pyo.NonNegativeReals,
            doc="P content of biomass, X_H, X_PAO, X_AUT, [kg P/kg COD]",
        )
        self.BOD5_factor = pyo.Param(
            ["raw", "effluent"],
            initialize={"raw": 0.65, "effluent": 0.25},
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Conversion factor for BOD5",
        )

        # Fix Vars that are treated as Params
        for v in self.component_objects(pyo.Var):
            v.fix()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "flow_vol": {"method": None},
                "pressure": {"method": None},
                "temperature": {"method": None},
                "conc_mass_comp": {"method": None},
            }
        )
        obj.define_custom_properties(
            {
                "VSS": {"method": "_VSS"},
                "ISS": {"method": "_ISS"},
                "TSS": {"method": "_TSS"},
                "COD": {"method": "_COD"},
                "TKN": {"method": "_TKN"},
                "SNOX": {"method": "_SNOX"},
                "BOD5": {"method": "_BOD5"},
                "SP_organic": {"method": "_SP_organic"},
                "SP_inorganic": {"method": "_SP_inorganic"},
            }
        )
        obj.add_default_units(
            {
                "time": pyo.units.s,
                "length": pyo.units.m,
                "mass": pyo.units.kg,
                "amount": pyo.units.kmol,
                "temperature": pyo.units.K,
            }
        )


class ModifiedASM2dPropertiesScaler(CustomScalerBase):
    """
    Scaler for the Activated Sludge Model No.2d property package.
    Flow and temperature are scaled by the default value (if no user input provided), and
    pressure is scaled assuming an order of magnitude of 1e5 Pa.
    """

    UNIT_SCALING_FACTORS = {
        # "QuantityName: (reference units, scaling factor)
        "Pressure": (pyo.units.Pa, 1e-5),
    }

    DEFAULT_SCALING_FACTORS = {
        "flow_vol": 1e1,
        "temperature": 1e-2,
    }

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        self.scale_variable_by_default(model.temperature, overwrite=overwrite)
        self.scale_variable_by_default(model.flow_vol, overwrite=overwrite)
        self.scale_variable_by_units(model.pressure, overwrite=overwrite)

    # There are currently no constraints in this model
    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        pass


class _ModifiedASM2dStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    default_scaler = ModifiedASM2dPropertiesScaler

    def initialize(
        self,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for property package.

        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:
            flow_vol : value at which to initialize total volumetric flow (default=None)
            conc_mass_comp : value at which to initialize component concentrations (default=None)
            pressure : value at which to initialize pressure (default=None)
            temperature : value at which to initialize temperature (default=None)
            outlvl : sets output level of initialization routine
            state_vars_fixed: Flag to denote if state vars have already been fixed.
                              True - states have already been fixed and
                              initialization does not need to worry
                              about fixing and unfixing variables.
                              False - states have not been fixed. The state
                              block will deal with fixing/unfixing.
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         True - states variables are not unfixed, and
                         a dict of returned containing flags for
                         which states were fixed during initialization.
                         False - state variables are unfixed after
                         initialization by calling the release_state method.

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")

        if state_vars_fixed is False:
            # Fix state variables if not already fixed
            flags = fix_state_vars(self, state_args)

        else:
            # Check when the state vars are fixed already result in dof 0
            for k in self.keys():
                if degrees_of_freedom(self[k]) != 0:
                    raise Exception(
                        "State vars fixed but degrees of freedom "
                        "for state block is not zero during "
                        "initialization."
                    )

        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)

        init_log.info("Initialization Complete.")

    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        """
        Method to relase state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")

        if flags is None:
            return
        # Unfix state variables
        revert_state_vars(self, flags)
        init_log.info("State Released.")


@declare_process_block_class(
    "ModifiedASM2dStateBlock", block_class=_ModifiedASM2dStateBlock
)
class ModifiedASM2dStateBlockData(StateBlockData):
    """
    StateBlock for calculating thermophysical proeprties associated with the ASM2d
    reaction system.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super().build()

        # Create state variables
        self.flow_vol = pyo.Var(
            initialize=1.0,
            domain=pyo.NonNegativeReals,
            doc="Total volumetric flowrate",
            units=pyo.units.m**3 / pyo.units.s,
        )
        self.pressure = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=101325.0,
            bounds=(1e3, 1e6),
            doc="Pressure",
            units=pyo.units.Pa,
        )
        self.temperature = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=298.15,
            bounds=(273.15, 323.15),
            doc="Temperature",
            units=pyo.units.K,
        )
        self.conc_mass_comp = pyo.Var(
            self.params.solute_set,
            domain=pyo.NonNegativeReals,
            initialize=0.1,
            doc="Component mass concentrations",
            units=pyo.units.kg / pyo.units.m**3,
        )

        # TODO: X_SRB not included yet in biomass term summation
        def _VSS(self):
            vss = (
                self.conc_mass_comp["X_I"] / self.params.CODtoVSS_XI
                + self.conc_mass_comp["X_S"] / self.params.CODtoVSS_XS
                + (
                    self.conc_mass_comp["X_H"]
                    + self.conc_mass_comp["X_PAO"]
                    + self.conc_mass_comp["X_AUT"]
                )
                / self.params.CODtoVSS_XBM
                + self.conc_mass_comp["X_PHA"] / self.params.CODtoVSS_XPHA
            )
            return vss

        self.VSS = pyo.Expression(rule=_VSS, doc="Volatile suspended solids")

        def _ISS(self):
            iss = (
                self.params.f_ISS_BM
                * (
                    self.conc_mass_comp["X_H"]
                    + self.conc_mass_comp["X_PAO"]
                    + self.conc_mass_comp["X_AUT"]
                )
                / self.params.CODtoVSS_XBM
                + self.params.ISS_P * self.conc_mass_comp["X_PP"]
            )
            return iss

        self.ISS = pyo.Expression(rule=_ISS, doc="Inorganic suspended solids")

        def _TSS(self):
            tss = self.VSS + self.ISS
            return tss

        self.TSS = pyo.Expression(rule=_TSS, doc="Total suspended solids")

        def _COD(self):
            cod = (
                self.conc_mass_comp["S_F"]
                + self.conc_mass_comp["S_A"]
                + self.conc_mass_comp["S_I"]
                + self.conc_mass_comp["X_I"]
                + self.conc_mass_comp["X_S"]
                + self.conc_mass_comp["X_H"]
                + self.conc_mass_comp["X_PAO"]
                + self.conc_mass_comp["X_PHA"]
                + self.conc_mass_comp["X_AUT"]
            )
            return cod

        self.COD = pyo.Expression(rule=_COD, doc="Chemical oxygen demand")

        def _TKN(self):
            tkn = (
                self.conc_mass_comp["S_NH4"]
                + self.params.i_NSF * self.conc_mass_comp["S_F"]
                + self.params.i_NSI * self.conc_mass_comp["S_I"]
                + self.params.i_NXI * self.conc_mass_comp["X_I"]
                + self.params.i_NXS * self.conc_mass_comp["X_S"]
                + self.params.i_NBM
                * (
                    self.conc_mass_comp["X_H"]
                    + self.conc_mass_comp["X_PAO"]
                    + self.conc_mass_comp["X_AUT"]
                )
            )
            return tkn

        self.TKN = pyo.Expression(rule=_TKN, doc="Kjeldahl nitrogen")

        def _SNOX(self):
            snox = self.conc_mass_comp["S_NO3"]

            return snox

        self.SNOX = pyo.Expression(rule=_SNOX, doc="Nitrogen oxide")

        def _BOD5(self, i):
            bod5 = (
                self.conc_mass_comp["S_F"]
                + self.conc_mass_comp["S_A"]
                + (1 - self.params.f_SI) * self.conc_mass_comp["X_S"]
                + (1 - self.params.f_XIH) * self.conc_mass_comp["X_H"]
                + (1 - self.params.f_XIP)
                * (self.conc_mass_comp["X_PAO"] + self.conc_mass_comp["X_PHA"])
                + (1 - self.params.f_XIA) * self.conc_mass_comp["X_AUT"]
            )

            return self.params.BOD5_factor[i] * bod5

        self.BOD5 = pyo.Expression(
            ["raw", "effluent"], rule=_BOD5, doc="Five-day biological oxygen demand"
        )

        def _SP_organic(self):
            sp_organic = (
                self.conc_mass_comp["X_PP"]
                + self.params.i_PSF * self.conc_mass_comp["S_F"]
                + self.params.i_PSI * self.conc_mass_comp["S_I"]
                + self.params.i_PXI * self.conc_mass_comp["X_I"]
                + self.params.i_PXS * self.conc_mass_comp["X_S"]
                + self.params.i_PBM
                * (
                    self.conc_mass_comp["X_H"]
                    + self.conc_mass_comp["X_PAO"]
                    + self.conc_mass_comp["X_AUT"]
                )
            )
            return sp_organic

        self.SP_organic = pyo.Expression(rule=_SP_organic, doc="Organic phosphorus")

        def _SP_inorganic(self):
            sp_inorganic = self.conc_mass_comp["S_PO4"]
            return sp_inorganic

        self.SP_inorganic = pyo.Expression(
            rule=_SP_inorganic, doc="Inorganic phosphorus"
        )

    def get_material_flow_terms(self, p, j):
        if j == "H2O":
            return self.flow_vol * self.params.dens_mass
        else:
            return self.flow_vol * self.conc_mass_comp[j]

    def get_enthalpy_flow_terms(self, p):
        return (
            self.flow_vol
            * self.params.dens_mass
            * self.params.cp_mass
            * (self.temperature - self.params.temperature_ref)
        )

    def get_material_density_terms(self, p, j):
        if j == "H2O":
            return self.params.dens_mass
        else:
            return self.conc_mass_comp[j]

    def get_energy_density_terms(self, p):
        return (
            self.params.dens_mass
            * self.params.cp_mass
            * (self.temperature - self.params.temperature_ref)
        )

    def default_material_balance_type(self):
        return MaterialBalanceType.componentPhase

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(self):
        return {
            "flow_vol": self.flow_vol,
            "conc_mass_comp": self.conc_mass_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    def define_display_vars(self):
        return {
            "Volumetric Flowrate": self.flow_vol,
            "Mass Concentration": self.conc_mass_comp,
            "Temperature": self.temperature,
            "Pressure": self.pressure,
        }

    def get_material_flow_basis(self):
        return MaterialFlowBasis.mass

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Set,
    Var,
    check_optimal_termination,
    Param,
    Constraint,
    Suffix,
    units as pyunits,
    NonNegativeReals,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.solvers import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import InitializationError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from watertap.costing.unit_models.surrogate_crystallizer import (
    cost_surrogate_crystallizer,
)

_log = idaeslog.getLogger(__name__)

__author__ = "Oluwamayowa Amusat, Adam Atia"


@declare_process_block_class("SurrogateCrystallizer")
class SurrogateCrystallizerData(UnitModelBlockData):
    """
    ML-based crystallizer model
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. NF units do not support dynamic
    behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. NF units do not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "vapor_property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for vapor phase",
            doc="""Property parameter object used to define property calculations
    for the vapor phase,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "vapor_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing vapor phase properties",
            doc="""A ConfigBlock with arguments to be passed to vapor phase
property block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "solids_ions_dict",
        ConfigValue(
            default={},
            domain=dict,
            description="Specification of ion makeup of each solid in system. e.g., {'NaCl': {'Na_+': 1, 'Cl_-': -1} }",
            doc="""Solids makeup; keys are solid names, and values include subkeys of ion names and values of charge.""",
        ),
    )

    def build(self):

        # Call UnitModel.build to setup dynamics
        super().build()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)
        units_meta = self.config.property_package.get_metadata().get_derived_units

        self.solids_list = Set(initialize=self.config.solids_ions_dict.keys())

        # Add ion in solid ratios as parameters
        dict1 = dict()
        for m in self.solids_list:
            for p in self.config.property_package.ion_set:
                dict1[m, p] = float(
                    self.config.solids_ions_dict[m][p]
                    if p in self.config.solids_ions_dict[m]
                    else 0.0
                )
        self.mwc = Param(
            self.solids_list, self.config.property_package.ion_set, initialize=dict1
        )

        # Add other variables
        self.temperature_operating = Var(
            initialize=298.15,
            domain=NonNegativeReals,
            units=pyunits.K,
            bounds=(273, 1000),
            doc="Crystallizer operating temperature in K",
        )
        self.pressure_operating = Var(
            initialize=101325,
            domain=NonNegativeReals,
            units=pyunits.Pa,
            bounds=(0.001, 1e6),
            doc="Crystallizer pressure in Pa",
        )
        self.evaporation_percent = Var(
            initialize=50,
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            bounds=(10, 95),
            doc="Crystallizer percentage of water evaporation",
        )
        self.flow_mass_sol_comp_true = Var(
            self.config.property_package.ion_set,
            initialize=0,
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.s,
            doc="True species solid mass flow (kg)",
        )
        self.flow_mass_sol_comp_apparent = Var(
            self.solids_list,
            initialize=0,
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.s,
            doc="Apparent species solid mass flow (kg)",
        )
        self.flow_mass_sol_total = Var(
            initialize=0,
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.s,
            doc="Total solid flow",
        )
        self.flow_mass_liq_total = Var(
            initialize=0,
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.s,
            doc="Total liquid flow",
        )
        self.flow_mass_vap_total = Var(
            initialize=0,
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.s,
            doc="Total vapor flow",
        )

        self.heat_required = Var(
            initialize=1e5,
            domain=NonNegativeReals,
            units=pyunits.kW,
            bounds=(-5e6, 5e6),
            doc="Heat requirement for crystallizer (kW)",
        )

        # # Add state blocks for inlet and three outlets
        # Add inlet block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # inlet block is an inlet
        self.properties_in = self.config.property_package.state_block_class(
            self.flowsheet().config.time, doc="Material properties of inlet", **tmp_dict
        )

        # Add liquid outlet
        tmp_dict["defined_state"] = False  # outlet and waste block is not an inlet
        self.properties_out_liq = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of liquid outlet",
            **tmp_dict,
        )

        # Add vapor outlet
        tmp_dict2 = dict()
        tmp_dict2["has_phase_equilibrium"] = False
        tmp_dict2["parameters"] = self.config.vapor_property_package
        tmp_dict["defined_state"] = False
        self.properties_out_vapor = (
            self.config.vapor_property_package.state_block_class(
                self.flowsheet().config.time,
                doc="Material properties of vapor outlet",
                **tmp_dict2,
            )
        )

        # Add ports - oftentimes users interact with these rather than the state blocks
        self.add_port(name="inlet", block=self.properties_in)
        self.add_port(name="outlet", block=self.properties_out_liq)
        self.add_port(name="vapor", block=self.properties_out_vapor)

        # Add constraints
        # 1. Fix empty flows:

        for p in self.config.vapor_property_package.phase_list:
            if p == "Liq":
                self.properties_out_vapor[0].flow_mass_phase_comp[p, "H2O"].fix(0.0)

        # 2. Material balances
        @self.Constraint(
            self.config.property_package.component_list,
            doc="Mass balance for components",
        )
        def eq_mass_balance_constraints(b, j):
            if j != "H2O":
                return (
                    sum(
                        b.properties_in[0].flow_mass_phase_comp[p, j]
                        for p in self.config.property_package.phase_list
                    )
                    == sum(
                        b.properties_out_liq[0].flow_mass_phase_comp[p, j]
                        for p in self.config.property_package.phase_list
                    )
                    + b.flow_mass_sol_comp_true[j]
                )
            else:
                return sum(
                    b.properties_in[0].flow_mass_phase_comp[p, j]
                    for p in self.config.property_package.phase_list
                ) == sum(
                    b.properties_out_liq[0].flow_mass_phase_comp[p, j]
                    for p in self.config.property_package.phase_list
                ) + sum(
                    b.properties_out_vapor[0].flow_mass_phase_comp[p, j]
                    for p in self.config.vapor_property_package.phase_list
                )

        # 3. Temperature and pressure balances:
        @self.Constraint()
        def eq_T_con1(b):
            return self.temperature_operating == b.properties_out_liq[0].temperature

        @self.Constraint()
        def eq_T_con2(b):
            return self.temperature_operating == b.properties_out_vapor[0].temperature

        @self.Constraint()
        def eq_P_con1(b):
            return self.pressure_operating == b.properties_out_liq[0].pressure

        @self.Constraint()
        def eq_P_con2(b):
            return self.pressure_operating == b.properties_out_vapor[0].pressure

        @self.Constraint(
            self.config.property_package.component_list,
            doc="Liquid-Vapour phase water split",
        )
        def eq_water_balance_constraints(b, j):
            if j == "H2O":
                return (
                    sum(
                        b.properties_out_vapor[0].flow_mass_phase_comp[p, j]
                        for p in self.config.vapor_property_package.phase_list
                    )
                    == b.properties_in[0].flow_mass_phase_comp["Liq", j]
                    * b.evaporation_percent
                    / 100
                )
            else:
                return Constraint.Skip

        @self.Constraint(
            self.config.property_package.component_list,
            doc="Mass balance for components",
        )
        def eq_solid_amount(b, j):
            if j == "H2O":
                return Constraint.Skip
            else:
                return b.flow_mass_sol_comp_true[j] == sum(
                    (b.mwc[q, j] * b.properties_in[0].params.mw_comp[j])
                    / sum(
                        b.mwc[q, j] * b.properties_in[0].params.mw_comp[j]
                        for j in b.config.property_package.ion_set
                    )
                    * b.flow_mass_sol_comp_apparent[q]
                    for q in b.solids_list
                )

        # 5. Constraints calculating total flows for each outlet stream
        @self.Constraint(
            doc="Total solid flow constraint",
        )
        def eq_total_solids_constraint(b):
            return b.flow_mass_sol_total == sum(
                b.flow_mass_sol_comp_apparent[k] for k in b.solids_list
            )

        @self.Constraint(
            doc="Total liquid flow constraint",
        )
        def eq_total_liquid_constraint(b):
            return b.flow_mass_liq_total == sum(
                sum(
                    b.properties_out_liq[0].flow_mass_phase_comp[p, j]
                    for p in self.config.property_package.phase_list
                )
                for j in self.config.property_package.component_list
            )

        @self.Constraint(
            doc="Total vapor flow constraint",
        )
        def eq_total_vapor_constraint(b):
            return b.flow_mass_vap_total == sum(
                b.properties_out_vapor[0].flow_mass_phase_comp[p, "H2O"]
                for p in self.config.vapor_property_package.phase_list
            )

        # 6. Constraints for computing heat requirement
        @self.Constraint(
            doc="Energy balance",
        )
        def eq_energy_balance_constraint(b):
            return pyunits.convert(
                b.heat_required, to_units=pyunits.W
            ) + b.properties_in[0].enth_flow == b.properties_out_liq[0].enth_flow + sum(
                b.properties_out_vapor[0].enth_flow_phase[p]
                for p in self.config.vapor_property_package.phase_list
            )

    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize holdup block
        flags = self.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
            hold_state=True,
        )
        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Initialize other state blocks
        # Set state_args from inlet state
        if state_args is None:
            state_args = {}
            state_dict = self.properties_in[
                self.flowsheet().config.time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        self.properties_out_liq.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )

        state_args_vapor = {}
        state_dict_vapor = self.properties_out_vapor[
            self.flowsheet().config.time.first()
        ].define_port_members()

        for k in state_dict_vapor.keys():
            if state_dict_vapor[k].is_indexed():
                state_args_vapor[k] = {}
                for m in state_dict_vapor[k].keys():
                    state_args_vapor[k][m] = state_dict_vapor[k][m].value
            else:
                state_args_vapor[k] = state_dict_vapor[k].value
        for p in self.config.vapor_property_package.phase_list:
            if p == "Vap":
                state_args_vapor["flow_mass_phase_comp"][p, "H2O"] = state_args[
                    "flow_mass_phase_comp"
                ]["Liq", "H2O"]
            else:
                state_args_vapor["flow_mass_phase_comp"][p, "H2O"] = 0
        self.properties_out_vapor.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_vapor,
        )
        init_log.info_high("Initialization Step 2 Complete.")
        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)

        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))
        # ---------------------------------------------------------------------
        # Release Inlet state
        self.properties_in.release_state(flags, outlvl=outlvl)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))
        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Feed Inlet": self.inlet,
                "Liquid Outlet": self.outlet,
                "Vapor Outlet": self.vapor,
            },
            time_point=time_point,
        )

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Operating Temperature (K)"] = self.temperature_operating
        var_dict["Vapor Pressure (Pa)"] = self.pressure_operating
        var_dict["Total solids at outlet (Kg)"] = self.flow_mass_sol_total
        var_dict["Total liquid flow at outlet (Kg)"] = self.flow_mass_liq_total
        var_dict["Total water vapor flow at outlet (Kg)"] = self.flow_mass_vap_total
        var_dict["Heat requirement"] = self.heat_required
        return {"vars": var_dict}

    def calculate_scaling_factors(self):

        super().calculate_scaling_factors()

        iscale.set_scaling_factor(
            self.temperature_operating,
            iscale.get_scaling_factor(self.properties_in[0].temperature),
        )
        iscale.set_scaling_factor(self.pressure_operating, 1e-3)
        iscale.set_scaling_factor(
            self.heat_required,
            iscale.get_scaling_factor(
                self.properties_in[0].flow_mass_phase_comp["Liq", "H2O"]
            )
            * iscale.get_scaling_factor(self.properties_in[0].enth_mass_phase["Liq"])
            * 1e3,
        )

        # transforming constraints
        for ind, c in self.eq_T_con1.items():
            sf = iscale.get_scaling_factor(self.properties_in[0].temperature)
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_T_con2.items():
            sf = iscale.get_scaling_factor(self.properties_in[0].temperature)
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_P_con1.items():
            sf = iscale.get_scaling_factor(self.properties_in[0].pressure)
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_P_con2.items():
            sf = iscale.get_scaling_factor(self.properties_in[0].pressure)
            iscale.constraint_scaling_transform(c, sf)

        for j, c in self.eq_mass_balance_constraints.items():
            sf = iscale.get_scaling_factor(
                self.properties_in[0].flow_mass_phase_comp["Liq", j]
            )
            iscale.constraint_scaling_transform(c, sf)

        for j, c in self.eq_water_balance_constraints.items():
            sf = iscale.get_scaling_factor(
                self.properties_in[0].flow_mass_phase_comp["Liq", "H2O"]
            )
            iscale.constraint_scaling_transform(c, sf)

        for j, c in self.eq_energy_balance_constraint.items():
            sf = iscale.get_scaling_factor(
                self.properties_in[0].flow_mass_phase_comp["Liq", "H2O"]
            ) * iscale.get_scaling_factor(self.properties_in[0].enth_mass_phase["Liq"])
            iscale.constraint_scaling_transform(c, sf)

    @property
    def default_costing_method(self):
        return cost_surrogate_crystallizer



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################


from pyomo.environ import (
    Var,
    Constraint,
    units as pyunits,
)

__author__ = "Alexander V. Dudchenko"


def cost_unit(
    costing_block,
    separator_unit,
    base_cost=0,
    mass_base_cost=0,
    additive_cost=0,
    species_cost=0,
    opt_name=None,
    valorizer_costing=False,
):
    separator_unit.base_cost = Var(
        initialize=base_cost, units=costing_block.base_currency / pyunits.m**3
    )
    separator_unit.additive_cost = Var(
        initialize=additive_cost, units=costing_block.base_currency / pyunits.kg
    )
    separator_unit.mass_base_cost = Var(
        initialize=mass_base_cost, units=costing_block.base_currency / pyunits.kg
    )
    if valorizer_costing == False:
        separator_unit.separation_cost = Var(
            list(separator_unit.component_removal_percent.keys()),
            initialize=species_cost,
            units=costing_block.base_currency / pyunits.kg,
        )

        separator_unit.separation_cost.fix()
        ion_multiplier = 1
        ion_cost = separator_unit.separation_cost

    else:
        separator_unit.product_value = Var(
            list(separator_unit.component_removal_percent.keys()),
            initialize=species_cost,
            units=costing_block.base_currency / pyunits.kg,
        )
        separator_unit.product_value.fix()
        ion_multiplier = -1
        ion_cost = separator_unit.product_value
    separator_unit.product_properties[0].flow_mass_phase_comp[...]
    separator_unit.base_cost.fix()
    separator_unit.additive_cost.fix()
    separator_unit.mass_base_cost.fix()
    separator_unit.annual_cost = Var(
        initialize=1, units=costing_block.base_currency / pyunits.year
    )
    separator_unit.feed_LCOW = Var(
        initialize=1, units=costing_block.base_currency / pyunits.m**3
    )
    separator_unit.annual_treatment_cost = Var(
        initialize=1, units=costing_block.base_currency / pyunits.year
    )
    separator_unit.annual_ion_separation_cost = Var(
        initialize=1, units=costing_block.base_currency / pyunits.year
    )

    separator_unit.ion_removal_LCOW = Var(
        initialize=1, units=costing_block.base_currency / pyunits.m**3
    )
    separator_unit.treatment_LCOW = Var(
        initialize=1, units=costing_block.base_currency / pyunits.m**3
    )
    separator_unit.annual_treatment_cost_eq = Constraint(
        expr=separator_unit.annual_treatment_cost
        == separator_unit.base_cost
        * pyunits.convert(
            separator_unit.separator_unit.properties_in[0].flow_vol_phase["Liq"],
            to_units=pyunits.m**3 / pyunits.year,
        )
        + separator_unit.additive_cost
        * pyunits.convert(
            separator_unit.additive_mass_flow,
            to_units=pyunits.kg / pyunits.year,
        )
        + separator_unit.mass_base_cost
        * sum(
            [
                pyunits.convert(
                    separator_unit.product_properties[0].flow_mass_phase_comp[
                        "Liq", ion
                    ],
                    to_units=pyunits.kg / pyunits.year,
                )
                for ion in ion_cost.keys()
            ]
        )
    )
    separator_unit.annual_ion_separation_cost_eq = Constraint(
        expr=separator_unit.annual_ion_separation_cost
        == sum(
            [
                ion_multiplier
                * ion_cost[ion]
                * pyunits.convert(
                    separator_unit.product_properties[0].flow_mass_phase_comp[
                        "Liq", ion
                    ],
                    to_units=pyunits.kg / pyunits.year,
                )
                for ion in ion_cost.keys()
            ]
        )
    )
    separator_unit.annual_cost_eq = Constraint(
        expr=separator_unit.annual_cost
        == separator_unit.annual_ion_separation_cost
        + separator_unit.annual_treatment_cost
    )

    separator_unit.feed_LCOW_eq = Constraint(
        expr=separator_unit.feed_LCOW
        == separator_unit.annual_cost
        / pyunits.convert(
            separator_unit.separator_unit.properties_in[0].flow_vol_phase["Liq"],
            to_units=pyunits.m**3 / pyunits.year,
        )
    )
    separator_unit.ion_removal_LCOW_eq = Constraint(
        expr=separator_unit.ion_removal_LCOW
        == ion_multiplier
        * separator_unit.annual_ion_separation_cost
        / pyunits.convert(
            separator_unit.separator_unit.properties_in[0].flow_vol_phase["Liq"],
            to_units=pyunits.m**3 / pyunits.year,
        )
    )

    separator_unit.water_treatment_LCOW_eq = Constraint(
        expr=separator_unit.treatment_LCOW
        == separator_unit.annual_treatment_cost
        / pyunits.convert(
            separator_unit.separator_unit.properties_in[0].flow_vol_phase["Liq"],
            to_units=pyunits.m**3 / pyunits.year,
        )
    )

    if valorizer_costing:
        costing_block.annual_costs.append(
            {
                "name": f"{opt_name} cost",
                "annual_cost": separator_unit.annual_treatment_cost,
            }
        )
        costing_block.annual_costs.append(
            {
                "name": f"{opt_name} revenue",
                "annual_cost": separator_unit.annual_ion_separation_cost,
            }
        )
    else:
        costing_block.annual_costs.append(
            {"name": opt_name, "annual_cost": separator_unit.annual_cost}
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from idaes_flowsheet_processor.api import FlowsheetInterface, FlowsheetCategory
from watertap.flowsheets.generic_desalination_train import generic_train
from pyomo.environ import units as pyunits
from idaes.core.solvers import get_solver


__author__ = "Alexander V. Dudchenko"


def export_to_ui():
    return FlowsheetInterface(
        name="Generic treatment train",
        do_export=export_variables,
        do_build=build_flowsheet,
        do_solve=solve_flowsheet,
        get_diagram=get_diagram,
        requires_idaes_solver=True,
        category=FlowsheetCategory.wastewater,
    )


def export_variables(flowsheet=None, exports=None, build_options=None, **kwargs):

    # print("model", model)
    fs = flowsheet  # model.fs
    # --- Input data ---
    # Feed conditions
    desalination_recovery = "Desalination recovery"
    water_management_costs = f"Water management costs"
    optional_inputs = f"Optional inputs"
    optional_costs = f"Optional cost"
    optional_inputs = f"Optional inputs"
    valorizer_inputs = "Valorization"
    desal_unit_costs = "Desalination unit cost"
    system_outputs = "System outputs"
    exports.add(
        obj=fs.costing.LCOW,
        name="LCOW Total",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m^3",
        rounding=3,
        description="LCOW",
        is_input=False,
        input_category="LCOW",
        is_output=True,
        output_category="LCOW",
    )

    for unit in fs.costing.LCOW_unit:
        exports.add(
            obj=fs.costing.LCOW_unit[unit],
            name=f'LCOW {unit.replace("_", " ")}',
            ui_units=fs.costing.base_currency / pyunits.m**3,
            display_units="$/m^3",
            rounding=3,
            description="LCOW",
            is_input=False,
            input_category="LCOW",
            is_output=True,
            output_category="LCOW",
            chart_type="stacked_bar_with_net",
            chart_group="process_cost",
        )
    for proc in fs.process_order:
        block = proc["process_block"]
        process_name = proc["process_name"]
        process_name_nice = proc["process_name"].replace("_", " ")

        if proc["process_type"] == "desalter":
            exports.add(
                obj=block.desalter.base_cost,
                name=f"{process_name_nice} base cost",
                ui_units=fs.costing.base_currency / pyunits.m**3,
                display_units="$/m^3",
                rounding=3,
                description=water_management_costs,
                is_input=True,
                input_category=water_management_costs,
                is_output=True,
                output_category=water_management_costs,
            )
            exports.add(
                obj=block.desalter.water_recovery,
                name=f"{process_name_nice} water recovery",
                ui_units=pyunits.dimensionless,
                display_units="%",
                rounding=3,
                description=desalination_recovery,
                is_input=True,
                input_category=desalination_recovery,
                is_output=True,
                output_category=desalination_recovery,
            )
            exports.add(
                obj=block.desalter.LCOW,
                name=f"{process_name_nice} unit cost",
                ui_units=fs.costing.base_currency / pyunits.m**3,
                display_units="$/m^3",
                rounding=3,
                description=desal_unit_costs,
                is_input=False,
                input_category=desal_unit_costs,
                is_output=True,
                output_category=desal_unit_costs,
            )

        if proc["process_type"] == "separator" or proc["process_type"] == "valorizer":
            if proc["process_type"] == "separator":
                exports.add(
                    obj=block.separator.base_cost,
                    name=f"{process_name_nice} base cost",
                    ui_units=fs.costing.base_currency / pyunits.m**3,
                    display_units="$/m^3",
                    rounding=3,
                    description=water_management_costs,
                    is_input=True,
                    input_category=water_management_costs,
                    is_output=True,
                    output_category=water_management_costs,
                )
            if proc["process_type"] == "valorizer":
                for (phase, ion), obj in fs.feed.properties[
                    0
                ].conc_mass_phase_comp.items():

                    if ion == "X":
                        exports.add(
                            obj=obj,
                            name="Feed {}".format(ion),
                            ui_units=pyunits.mg / pyunits.L,
                            display_units="mg/L",
                            rounding=0,
                            description="{} concentration".format(ion),
                            is_input=True,
                            input_category=valorizer_inputs,
                            is_output=True,
                            output_category=valorizer_inputs,
                        )
                exports.add(
                    obj=block.separator.mass_base_cost,
                    name=f"{process_name_nice} base cost",
                    ui_units=fs.costing.base_currency / pyunits.kg,
                    display_units="$/kg",
                    rounding=3,
                    description=valorizer_inputs,
                    is_input=True,
                    input_category=valorizer_inputs,
                    is_output=True,
                    output_category=valorizer_inputs,
                )

                for ion in block.separator.product_value.keys():
                    if ion != "TDS" and ion != "H2O":
                        exports.add(
                            obj=block.separator.component_removal_percent[ion],
                            name=f"Valorizer {ion} recovery %",
                            ui_units=pyunits.dimensionless,
                            display_units="%",
                            rounding=3,
                            description=valorizer_inputs,
                            is_input=True,
                            input_category=valorizer_inputs,
                            is_output=True,
                            output_category=valorizer_inputs,
                        )
                        exports.add(
                            obj=block.separator.product_value[ion],
                            name=f"Valorizer {ion} value",
                            ui_units=fs.costing.base_currency / pyunits.kg,
                            display_units="$/kg",
                            rounding=3,
                            description=valorizer_inputs,
                            is_input=True,
                            input_category=valorizer_inputs,
                            is_output=True,
                            output_category=valorizer_inputs,
                        )
                exports.add(
                    obj=block.separator.treatment_LCOW,
                    name=f"Valorizer unit cost",
                    ui_units=fs.costing.base_currency / pyunits.m**3,
                    display_units="$/m^3",
                    rounding=3,
                    description=valorizer_inputs,
                    is_input=False,
                    input_category=valorizer_inputs,
                    is_output=True,
                    output_category=valorizer_inputs,
                )
                exports.add(
                    obj=block.separator.ion_removal_LCOW,
                    name=f"Valorizer levelized revenue (LROW)",
                    ui_units=fs.costing.base_currency / pyunits.m**3,
                    display_units="$/m^3",
                    rounding=3,
                    description=valorizer_inputs,
                    is_input=False,
                    input_category=valorizer_inputs,
                    is_output=True,
                    output_category=valorizer_inputs,
                )
                for (phase, ion), obj in block.product_mass_flow_feed_basis.items():
                    if ion != "H2O" and ion != "TDS":
                        exports.add(
                            obj=obj,
                            name="Valorizer {} production".format(ion),
                            ui_units=pyunits.kg / pyunits.m**3,
                            display_units="kg/m^3  of feed",
                            rounding=3,
                            description="{} mass flow".format(ion),
                            is_input=False,
                            input_category=valorizer_inputs,
                            is_output=True,
                            output_category=valorizer_inputs,
                        )

    exports.add(
        obj=fs.feed.base_cost,
        name="Feed source cost",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m^3",
        rounding=3,
        description=optional_costs,
        is_input=True,
        input_category=optional_costs,
        is_output=True,
        output_category=optional_costs,
    )
    exports.add(
        obj=fs.product.base_cost,
        name="Product distribution cost",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m^3",
        rounding=3,
        description=optional_costs,
        is_input=True,
        input_category=optional_costs,
        is_output=True,
        output_category=optional_costs,
    )
    for proc in fs.process_order:
        block = proc["process_block"]
        process_name = proc["process_name"]
        process_name_nice = proc["process_name"].replace("_", " ")
        if proc["process_type"] == "desalter":
            exports.add(
                obj=block.desalter.recovery_cost,
                name=f"{process_name_nice} rate cost LCOW/WR",
                ui_units=fs.costing.base_currency / pyunits.m**3,
                display_units="$/m^3/%",
                rounding=3,
                description=optional_costs,
                is_input=True,
                input_category=optional_costs,
                is_output=True,
                output_category=optional_costs,
            )
            exports.add(
                obj=block.desalter.recovery_cost_offset,
                name=f"{process_name_nice} rate cost offset ",
                ui_units=pyunits.dimensionless,
                display_units="%",
                rounding=2,
                description=optional_costs,
                is_input=True,
                input_category=optional_costs,
                is_output=False,
                output_category=optional_costs,
            )
    exports.add(
        obj=fs.disposal.base_cost,
        name="Disposal cost",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m^3",
        rounding=3,
        description=water_management_costs,
        is_input=True,
        input_category=water_management_costs,
        is_output=True,
        output_category=water_management_costs,
    )

    for (phase, ion), obj in fs.feed.properties[0].conc_mass_phase_comp.items():
        if ion != "H2O" and ion != "X":
            exports.add(
                obj=obj,
                name="Feed {}".format(ion),
                ui_units=pyunits.mg / pyunits.L,
                display_units="mg/L",
                rounding=0,
                description="{} concentration".format(ion),
                is_input=True,
                input_category=optional_inputs,
                is_output=True,
                output_category=optional_inputs,
            )

    exports.add(
        obj=fs.water_recovery,
        name="System water recovery",
        ui_units=pyunits.dimensionless,
        display_units="%",
        rounding=3,
        description=optional_inputs,
        is_input=True,
        input_category=optional_inputs,
        is_output=True,
        output_category=optional_inputs,
    )

    for proc in fs.process_order:
        block = proc["process_block"]
        process_name = proc["process_name"]
        process_name_nice = proc["process_name"].replace("_", " ")
        if proc["process_type"] == "desalter":
            if process_name != "Desal_3":
                exports.add(
                    obj=block.desalter.brine_solids_concentration,
                    name=f"{process_name_nice} outlet TDS",
                    ui_units=pyunits.g / pyunits.L,
                    display_units="g/L",
                    rounding=3,
                    description=optional_inputs,
                    is_input=True,
                    input_category=optional_inputs,
                    is_output=True,
                    output_category=optional_inputs,
                )
            else:
                exports.add(
                    obj=block.desalter.brine_solids_concentration,
                    name=f"{process_name_nice} sludge TDS",
                    ui_units=pyunits.g / pyunits.L,
                    display_units="g/L",
                    rounding=3,
                    description=optional_inputs,
                    is_input=False,
                    input_category=optional_inputs,
                    is_output=True,
                    output_category=optional_inputs,
                )
                exports.add(
                    obj=block.desalter.brine_water_mass_percent,
                    name=f"{process_name_nice} sludge water content",
                    ui_units=pyunits.dimensionless,
                    display_units="%",
                    rounding=2,
                    description=optional_inputs,
                    is_input=True,
                    input_category=optional_inputs,
                    is_output=True,
                    output_category=optional_inputs,
                )

    for (phase, ion), obj in fs.disposal.properties[0].conc_mass_phase_comp.items():
        if ion != "H2O" and ion != "X":
            exports.add(
                obj=obj,
                name="Waste {} concentration".format(ion),
                ui_units=pyunits.g / pyunits.L,
                display_units="g/L",
                rounding=3,
                description=system_outputs,
                is_input=False,
                input_category=system_outputs,
                is_output=True,
                output_category=system_outputs,
            )
            if ion == "TDS":
                exports.add(
                    obj=fs.disposal_mass_flow_feed_basis[phase, ion],
                    name="Waste solids production",  # .format(ion),
                    ui_units=pyunits.kg / pyunits.m**3,
                    display_units="kg/m^3 of feed",
                    rounding=3,
                    description=system_outputs,
                    is_input=False,
                    input_category=system_outputs,
                    is_output=True,
                    output_category=system_outputs,
                )


def build_flowsheet(build_options=None, **kwargs):
    # build and solve initial flowsheet
    solver = get_solver()
    m = generic_train.build()
    generic_train.initialize(m, solver)
    return m


def get_diagram(build_options):
    return "fig_with_costs.png"


def solve_flowsheet(flowsheet):
    solver = get_solver()
    results = generic_train.solve(flowsheet, solver)
    return results



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest

from pyomo.util.check_units import assert_units_consistent
import pyomo.environ as pyo
import idaes.core as idc

from watertap.costing.watertap_costing_package import WaterTAPCosting
import watertap.flowsheets.lsrro.lsrro as lsrro


@pytest.mark.component
def test_watertap_costing_package():
    m = pyo.ConcreteModel()
    m.fs = idc.FlowsheetBlock(dynamic=False)

    m.fs.costing = WaterTAPCosting()

    m.fs.electricity = pyo.Var(units=pyo.units.kW, initialize=1)

    m.fs.costing.cost_flow(m.fs.electricity, "electricity")

    assert "foo" not in m.fs.costing.flow_types
    with pytest.raises(
        ValueError,
        match="foo is not a recognized flow type. Please check "
        "your spelling and that the flow type has been registered with"
        " the FlowsheetCostingBlock.",
    ):
        m.fs.costing.cost_flow(m.fs.electricity, "foo")

    m.fs.costing.foo_cost = foo_cost = pyo.Var(
        initialize=42,
        doc="foo",
        units=pyo.units.USD_2020 / pyo.units.m / pyo.units.second,
    )

    m.fs.costing.register_flow_type("foo", m.fs.costing.foo_cost)

    # make sure the component was not replaced
    # by register_defined_flow
    assert foo_cost is m.fs.costing.foo_cost

    m.fs.foo = pyo.Var(units=pyo.units.m, initialize=10)

    m.fs.costing.cost_flow(m.fs.foo, "foo")

    m.fs.costing.bar_base_cost = pyo.Var(
        initialize=0.42,
        doc="bar",
        units=pyo.units.USD_2020 / pyo.units.g / pyo.units.hour,
    )
    m.fs.costing.bar_purity = pyo.Param(
        initialize=0.50, doc="bar purity", units=pyo.units.dimensionless
    )

    m.fs.costing.register_flow_type(
        "bar", m.fs.costing.bar_base_cost * m.fs.costing.bar_purity
    )

    bar_cost = m.fs.costing.bar_cost
    assert isinstance(bar_cost, pyo.Expression)
    assert pyo.value(bar_cost) == 0.21

    m.fs.costing.bar_base_cost.value = 1.5
    assert pyo.value(bar_cost) == 0.75

    m.fs.costing.baz_cost = pyo.Var(initialize=5)

    with pytest.raises(
        RuntimeError,
        match="Component baz_cost already exists on fs.costing but is not 42",
    ):
        m.fs.costing.register_flow_type(
            "baz", 42 * pyo.units.USD_2020 / pyo.units.m**2 / pyo.units.day
        )

    m.fs.costing.flow_types.remove("baz")

    m.fs.costing.register_flow_type(
        "ham", 42 * pyo.units.USD_2021 / pyo.units.kg / pyo.units.minute
    )

    assert isinstance(m.fs.costing.ham_cost, pyo.Var)

    m.fs.costing.cost_process()
    # no error, wacc, plant_lifetime fixed
    m.fs.costing.initialize()

    m.fs.costing.capital_recovery_factor.fix()
    with pytest.raises(
        RuntimeError,
        match="Exactly two of the variables fs.costing.plant_lifetime, "
        "fs.costing.wacc, fs.costing.capital_recovery_factor should be "
        "fixed and the other unfixed.",
    ):
        # error, capital_recovery_factor,  wacc, plant_lifetime all fixed
        m.fs.costing.initialize()
    m.fs.costing.wacc.unfix()

    # no error
    m.fs.costing.initialize()

    m.fs.costing.capital_recovery_factor.unfix()
    with pytest.raises(
        RuntimeError,
        match="Exactly two of the variables fs.costing.plant_lifetime, "
        "fs.costing.wacc, fs.costing.capital_recovery_factor should be "
        "fixed and the other unfixed.",
    ):
        # error, capital_recovery_factor, wacc, unfixed
        m.fs.costing.initialize()

    m.fs.costing.plant_lifetime.unfix()
    with pytest.raises(
        RuntimeError,
        match="Exactly two of the variables fs.costing.plant_lifetime, "
        "fs.costing.wacc, fs.costing.capital_recovery_factor should be "
        "fixed and the other unfixed.",
    ):
        # error, capital_recovery_factor, wacc, and plant_lifetime unfixed
        m.fs.costing.initialize()

    m.fs.costing.wacc.fix()
    m.fs.costing.capital_recovery_factor.fix()
    # no error, wacc, capital_recovery_factor fixed
    m.fs.costing.initialize()


@pytest.mark.component
def test_breakdowns():
    m = lsrro.build()

    m.fs.BoosterPumps[:].control_volume.work[0.0].value = 42e6
    m.fs.EnergyRecoveryDevices[:].control_volume.work[0.0].value = -42e6

    m.fs.costing.add_specific_electrical_carbon_intensity(
        m.fs.product.properties[0].flow_vol
    )

    assert_units_consistent(m)

    m.fs.costing.initialize()

    total_LCOW = pyo.value(m.fs.costing.LCOW)

    summed_aggregates = pyo.value(
        sum(m.fs.costing.LCOW_aggregate_direct_capex.values())
        + sum(m.fs.costing.LCOW_aggregate_indirect_capex.values())
        + sum(m.fs.costing.LCOW_aggregate_fixed_opex.values())
        + sum(m.fs.costing.LCOW_aggregate_variable_opex.values())
        # electricity is counted both in the aggregate variable opex
        # per unit and per flow, so it is double-counted in this sum
        - m.fs.costing.LCOW_aggregate_variable_opex["electricity"]
    )
    assert pytest.approx(total_LCOW) == summed_aggregates

    summed_components = pyo.value(
        sum(m.fs.costing.LCOW_component_direct_capex.values())
        + sum(m.fs.costing.LCOW_component_indirect_capex.values())
        + sum(m.fs.costing.LCOW_component_fixed_opex.values())
        + sum(m.fs.costing.LCOW_component_variable_opex.values())
    )
    assert pytest.approx(total_LCOW) == summed_components

    sec = pyo.value(m.fs.costing.specific_energy_consumption)
    summed_sec = pyo.value(
        sum(m.fs.costing.specific_energy_consumption_component.values())
    )
    assert pytest.approx(sec) == summed_sec

    seci = pyo.value(m.fs.costing.specific_electrical_carbon_intensity)
    summed_seci = pyo.value(
        sum(m.fs.costing.specific_electrical_carbon_intensity_component.values())
    )
    assert pytest.approx(seci) == summed_seci
